unit main;

{
Version

1.0.0  2011-02-19  Initial Version

}


interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, IdCoder, IdCoder3to4, IdCoderMIME, IdBaseComponent;

type
  TfrmMain = class(TForm)
    btnEncode: TButton;
    btnDecode: TButton;
    GroupBox1: TGroupBox;
    memClip: TMemo;
    Encode64: TIdEncoderMIME;
    Decode64: TIdDecoderMIME;
    dlgOpen: TOpenDialog;
    GroupBox2: TGroupBox;
    optFile: TRadioButton;
    optClip: TRadioButton;
    chkComp: TCheckBox;
    lblStatus: TLabel;
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure btnEncodeClick(Sender: TObject);
    procedure btnDecodeClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure GroupBox2MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  frmMain: TfrmMain;

implementation
{$R *.DFM}
uses Clipbrd, IdHashCRC, ZLib;

procedure CompressStream(inpStream, outStream: TStream);
var
  InpBuf, OutBuf: Pointer;
  InpBytes, OutBytes: Integer;
begin
  InpBuf := nil;
  OutBuf := nil;
  try
    GetMem(InpBuf, inpStream.Size);
    inpStream.Position := 0;
    InpBytes := inpStream.Read(InpBuf^, inpStream.Size);
    CompressBuf(InpBuf, InpBytes, OutBuf, OutBytes);
    outStream.Write(OutBuf^, OutBytes);
  finally
    if InpBuf <> nil then FreeMem(InpBuf);
    if OutBuf <> nil then FreeMem(OutBuf);
  end;
end;

procedure DecompressStream(inpStream, outStream: TStream);
var
  InpBuf, OutBuf: Pointer;
  OutBytes, sz: Integer;
begin
  InpBuf := nil;
  OutBuf := nil;
  sz     := inpStream.Size - inpStream.Position;
  if sz > 0 then
    try
      GetMem(InpBuf, sz);
      inpStream.Read(InpBuf^, sz);
      DecompressBuf(InpBuf, sz, 0, OutBuf, OutBytes);
      outStream.Write(OutBuf^, OutBytes);
    finally
      if InpBuf <> nil then FreeMem(InpBuf);
      if OutBuf <> nil then FreeMem(OutBuf);
    end;
  outStream.Position := 0;
end;


function IsNumeric(s : string): Boolean;
begin
  try
    StrToFloat(s);
    Result := TRUE;
  except
    on E: EConvertError do
    begin
      Result := FALSE;
    end;
  end;
end;

function StripChars(chars:string; text:string; bExcept:boolean): string;
  // Remove [all except] specified characters from text
  var
    i, n: integer;
    s: string;
begin
  n := length(text);
  s := '';
  for i := 1 to n do
  begin
    if isdelimiter(chars,text,i) = bExcept then
      s := s + text[i];
  end;
  result := s;
end;

function Wrap(text: string; cols: integer): string;
  var
    s: string;
    i: integer;
begin
  s := '';
  i := 1;
  while i < length(text) do
  begin
    if i > 1 then
      s := s + #13#10;
    s := s + copy(text,i,cols);
    i := i + cols;
  end;
  result := s;
end;

procedure ShowBusy(sMsg: string);
begin
  frmMain.lblStatus.Caption := sMsg;
  frmMain.optFile.Enabled := False;
  frmMain.optClip.Enabled := False;
  frmMain.chkComp.Enabled := False;
  frmMain.btnEncode.Enabled := False;
  frmMain.btnDecode.Enabled := False;
  frmMain.memClip.Enabled := False;
  Application.ProcessMessages;
end;

procedure ShowFree;
begin
  frmMain.optFile.Enabled := True;
  frmMain.optClip.Enabled := True;
  frmMain.chkComp.Enabled := True;
  frmMain.btnEncode.Enabled := True;
  frmMain.btnDecode.Enabled := True;
  frmMain.memClip.Enabled := True;
  frmMain.lblStatus.Caption := '';
  Application.ProcessMessages;
end;

procedure ShowError(s: string);
begin
  ShowFree;
  showmessage(s);
end;

procedure Encode();
  const Ver='1';
  var
    LenIn, LenOut: integer;
    strmIn, strmOut: TMemoryStream;
    s, Filename: string;
    sHeader, sBody, sChecksum, sOpts: string;
begin
  s := '';
  sOpts := '';
  if frmMain.optFile.Checked then
  begin
    frmMain.dlgOpen.Options := [ofHideReadOnly,ofFileMustExist,ofEnableSizing];
    if frmMain.dlgOpen.Execute then
    begin
      Filename := StripChars('()!',ExtractFileName(frmMain.dlgOpen.FileName),false); //make safe
      ShowBusy('Encoding "'+Filename+'"...');
      strmIn := TMemoryStream.Create;
      strmIn.LoadFromFile(frmMain.dlgOpen.FileName);
    end else
    begin
      exit;
    end;
  end else
  begin
    ShowBusy('Encoding clipboard contents...');
    s := Clipboard.AsText;
    strmIn := TMemoryStream.Create;
    strmIn.SetSize(length(s));
    strmIn.WriteBuffer(s[1],length(s));
    strmIn.Seek(0,soFromBeginning);
    Filename := '';
  end;
  LenIn := strmIn.size;
  strmOut := TMemoryStream.Create;
  if frmMain.chkComp.Checked then
  begin
    //Compress
    CompressStream(strmIn,strmOut);
    strmOut.Seek(0,soFromBeginning);
    If strmIn.size > strmOut.size then
    begin
      sOpts := sOpts + 'c';
    end else
    begin
      //compression makes it bigger!  Use Compression-substitute
      strmIn.Seek(0,soFromBeginning);
      strmOut.CopyFrom(strmIn,length(s));
      strmOut.Seek(0,soFromBeginning);
    end;
  end else
  begin
    //Compression-substitute
    strmOut.CopyFrom(strmIn,length(s));
    strmOut.Seek(0,soFromBeginning);
  end;

  sBody := frmMain.Encode64.Encode(strmOut,MaxInt);
  strmIn.Free;
  strmOut.Free;


  with TIdHashCRC32.Create do
  begin
    sChecksum := inttostr(HashValue(sBody));
    free;
    //Clipboard.SetTextBuf(pchar(sChecksum+#13#10+sBody));
    //showmessage('Harvest Clipboard!');
  end;
  sHeader := 'kwikx' + Ver + '!' + Filename + '!' + sOpts + '!' + sChecksum + '!';
  sBody := Wrap(sBody,70);
  s := StringOfChar('(',8) + sHeader + #13#10 + sBody + #13#10 + StringOfChar(')',8); //add header/footer
  if length(s) = 0 then
  begin
    frmMain.lblStatus.Caption := 'Nothing to encode!';
    ShowFree;
  end else
  begin
    Clipboard.SetTextBuf(pchar(s));
    frmMain.memClip.Text := s;
    LenOut := length(s);
    ShowFree;
    frmMain.lblStatus.Caption := 'Encoded '+inttostr(LenIn)+' --> '+inttostr(LenOut)+' bytes ('+inttostr((100*LenOut) div LenIn)+'%).';
  end;
end;

{
Encoding format:
((((((((kwikx1!FILENAME!OPTIONS!CHECKSUM!BASE64_PAYLOAD))))))))
}

procedure Decode();
  var
    LenIn, LenOut: integer;
    strmIn, strmOut: TMemoryStream;
    i: integer;
    s, Filename: string;
    sVer, sBody, sChecksum, sCalcChecksum, sOpts: string;
begin
  ShowBusy('Decoding clipboard contents...');
  s := Clipboard.AsText;
  LenIn := length(s);
  if LenIn = 0 then
  begin
    ShowError('Nothing on clipboard!');
    exit;
  end;
  if (s[1] <> '(') or (s[length(s)] <> ')') then
  begin
    ShowError('Clipboard text must begin with one or more "(" characters,'+#13#10+'and end with one or more ")" characters.');
    exit;
  end;
  i := lastdelimiter('(',s);
  s := copy(s, i+1, pos(')',s)-i-1);

  //Version
  i := pos('!',s);
  if i = 0 then
  begin
    ShowError('Couldn''t find version in header.');
    exit;
  end;
  if copy(s,1,5) <> 'kwikx' then
  begin
    ShowError('No "kwikx" in version field.');
    exit;
  end;
  sVer := copy(s,6,i-6);
  s := copy(s,i+1,maxint);
  if not isnumeric(sVer) then
  begin
    ShowError('Non-numeric version field.');
    exit;
  end;

  //Filename
  i := pos('!',s);
  if i = 0 then
  begin
    ShowError('Couldn''t find filename in header.');
    exit;
  end;
  Filename := copy(s,1,i-1);
  s := copy(s,i+1,maxint);

  //Options
  i := pos('!',s);
  if i = 0 then
  begin
    ShowError('Couldn''t find options in header.');
    exit;
  end;
  sOpts := copy(s,1,i-1);
  s := copy(s,i+1,maxint);

  //Checksum
  i := pos('!',s);
  if i = 0 then
  begin
    ShowError('Couldn''t find checksum in header.');
    exit;
  end;
  sChecksum := copy(s,1,i-1);
  s := copy(s,i+1,maxint);
  if not isnumeric(sChecksum) then
  begin
    ShowError('Non-numeric checksum field.');
    exit;
  end;

  //Body
  s := StripChars(GBase64CodeTable,s,true); //Remove anything that isn't a valid base64 character
  i := length(s) mod 4;
  if i > 0 then
    s := s + StringOfChar('=',4-i); //pad to 4-char boundary
  sBody := s;

  //Verify checksum
  with TIdHashCRC32.Create do
  begin
    sCalcChecksum := inttostr(HashValue(sBody));
    free;
  end;
  if sCalcChecksum <> sChecksum then
  begin
    ShowError('Checksums do not match.');
    //Clipboard.SetTextBuf(pchar(sChecksum+#13#10+sBody));
    //showmessage('Harvest Clipboard!');
    exit;
  end;


  strmIn := TMemoryStream.Create;
  strmOut := TMemoryStream.Create;
  frmMain.Decode64.DecodeToStream(sBody,strmIn);

  if pos('c',sOpts) > 0 then
  begin
    //Decompress
    strmIn.Seek(0,soFromBeginning);
    DecompressStream(strmIn,strmOut);
    strmOut.Seek(0,soFromBeginning);
  end else
  begin
    //Decompression-substitute
    strmIn.Seek(0,soFromBeginning);
    strmOut.CopyFrom(strmIn,strmIn.Size);
    strmOut.Seek(0,soFromBeginning);
  end;

  if frmMain.optFile.Checked then
  begin
    frmMain.dlgOpen.Options := [ofHideReadOnly,ofEnableSizing];
    frmMain.dlgOpen.FileName := Filename;
    ShowFree;
    if frmMain.dlgOpen.Execute then
    begin
      ShowBusy('Saving "'+ExtractFileName(frmMain.dlgOpen.FileName)+'"...');
      strmOut.SaveToFile(frmMain.dlgOpen.FileName);
    end;
  end else
  begin
    SetLength(s,strmOut.Size);
    strmOut.ReadBuffer(s[1],strmOut.Size);
    Clipboard.SetTextBuf(pchar(s));
    frmMain.memClip.Text := s;
  end;
  LenOut := strmOut.size;
  strmIn.Free;
  strmOut.Free;
  ShowFree;
  frmMain.lblStatus.Caption := 'Decoded '+inttostr(LenIn) + ' ('+inttostr((100*LenIn) div LenOut)+'%) --> '+inttostr(LenOut)+' bytes.';
end;

procedure UpdateMemo;
begin
  frmMain.memClip.clear;
  frmMain.memClip.PasteFromClipboard;
end;

procedure TfrmMain.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  UpdateMemo;
end;

procedure TfrmMain.btnEncodeClick(Sender: TObject);
begin
  Encode;
end;

procedure TfrmMain.btnDecodeClick(Sender: TObject);
begin
  Decode;
end;

procedure TfrmMain.FormActivate(Sender: TObject);
begin
  lblStatus.Caption := '';
  UpdateMemo;
end;

procedure TfrmMain.GroupBox2MouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
begin
  UpdateMemo;
end;

end.
